import copy
import typing
from dataclasses import dataclass
from datetime import timedelta
from pathlib import Path
from typing import Callable

import jinja2
from pydantic import BaseModel as PydanticBaseModel
from panther.throttling import Throttling

class JWTConfig:
    def __init__(
            self,
            key: str,
            algorithm: str = 'HS256',
            life_time: timedelta | int = timedelta(days=1),
            refresh_life_time: timedelta | int | None = None,
    ):
        self.key = key
        self.algorithm = algorithm
        self.life_time = life_time.total_seconds() if isinstance(life_time, timedelta) else life_time

        if refresh_life_time:
            if isinstance(refresh_life_time, timedelta):
                self.refresh_life_time = refresh_life_time.total_seconds()
            else:
                self.refresh_life_time = refresh_life_time
        else:
            self.refresh_life_time = self.life_time * 2


class QueryObservable:
    observers = []

    @classmethod
    def observe(cls, observer):
        cls.observers.append(observer)

    @classmethod
    def update(cls):
        for observer in cls.observers:
            observer._reload_bases(parent=config.QUERY_ENGINE)


@dataclass
class Config:
    BASE_DIR: Path
    MONITORING: bool
    LOG_QUERIES: bool
    DEFAULT_CACHE_EXP: timedelta | None
    THROTTLING: Throttling | None
    SECRET_KEY: bytes | None
    HTTP_MIDDLEWARES: list
    WS_MIDDLEWARES: list
    USER_MODEL: type[PydanticBaseModel] | None # type: type[panther.db.Model]
    AUTHENTICATION: type[PydanticBaseModel] | None
    WS_AUTHENTICATION: type[PydanticBaseModel] | None
    JWT_CONFIG: JWTConfig | None
    MODELS: list[type[PydanticBaseModel]]  # type: type[panther.db.Model]
    FLAT_URLS: dict
    URLS: dict
    WEBSOCKET_CONNECTIONS: typing.Callable | None
    BACKGROUND_TASKS: bool
    HAS_WS: bool
    STARTUPS: list[Callable]
    SHUTDOWNS: list[Callable]
    TIMEZONE: str
    TEMPLATES_DIR: str | list[str]
    JINJA_ENVIRONMENT: jinja2.Environment | None
    AUTO_REFORMAT: bool
    QUERY_ENGINE: typing.Callable | None
    DATABASE: typing.Callable | None

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        if key == 'QUERY_ENGINE' and value:
            QueryObservable.update()

    def __setitem__(self, key, value):
        setattr(self, key.upper(), value)

    def __getitem__(self, item):
        return getattr(self, item.upper())

    def refresh(self):
        # In some tests we need to `refresh` the `config` values
        for key, value in copy.deepcopy(default_configs).items():
            setattr(self, key, value)


default_configs = {
    'BASE_DIR': Path(),
    'MONITORING': False,
    'LOG_QUERIES': False,
    'DEFAULT_CACHE_EXP': None,
    'THROTTLING': None,
    'SECRET_KEY': None,
    'HTTP_MIDDLEWARES': [],
    'WS_MIDDLEWARES': [],
    'USER_MODEL': None,
    'AUTHENTICATION': None,
    'WS_AUTHENTICATION': None,
    'JWT_CONFIG': None,
    'MODELS': [],
    'FLAT_URLS': {},
    'URLS': {},
    'WEBSOCKET_CONNECTIONS': None,
    'BACKGROUND_TASKS': False,
    'HAS_WS': False,
    'STARTUPS': [],
    'SHUTDOWNS': [],
    'TIMEZONE': 'UTC',
    'TEMPLATES_DIR': '.',
    'JINJA_ENVIRONMENT': None,
    'AUTO_REFORMAT': False,
    'QUERY_ENGINE': None,
    'DATABASE': None,
}

config = Config(**copy.deepcopy(default_configs))



import asyncio
from dataclasses import dataclass
from http import cookies
from sys import version_info
from types import NoneType
from typing import Generator, AsyncGenerator, Any, Type, Literal

if version_info >= (3, 11):
    from typing import LiteralString
else:
    from typing import TypeVar

    LiteralString = TypeVar('LiteralString')

import orjson as json
from pydantic import BaseModel

from panther import status
from panther.configs import config
from panther._utils import to_async_generator
from panther.db.cursor import Cursor
from pantherdb import Cursor as PantherDBCursor
from panther.pagination import Pagination

ResponseDataTypes = list | tuple | set | Cursor | PantherDBCursor | dict | int | float | str | bool | bytes | NoneType | Type[BaseModel]
IterableDataTypes = list | tuple | set | Cursor | PantherDBCursor
StreamingDataTypes = Generator | AsyncGenerator


@dataclass(slots=True)
class Cookie:
    """
    path: [Optional] Indicates the path that must exist in the requested URL for the browser to send the Cookie header.
        Default is `/`
    domain: [Optional] Defines the host to which the cookie will be sent.
        Default is the host of the current document URL, not including subdomains.
    max_age: [Optional] Indicates the number of seconds until the cookie expires.
        A zero or negative number will expire the cookie immediately.
    secure: [Optional] Indicates that the cookie is sent to the server
        only when a request is made with the https: scheme (except on localhost)
    httponly: [Optional] Forbids JavaScript from accessing the cookie,
        for example, through the `Document.cookie` property.
    samesite: [Optional] Controls whether a cookie is sent with cross-site requests or not,
        `lax` is the default behavior if not specified.
    expires: [Deprecated] In HTTP version 1.1, `expires` was deprecated and replaced with the easier-to-use `max-age`
    """
    key: str
    value: str
    domain: str = None
    max_age: int = None
    secure: bool = False
    httponly: bool = False
    samesite: Literal['none', 'lax', 'strict'] = 'lax'
    path: str = '/'


class Response:
    content_type = 'application/json'

    def __init__(
        self,
        data: ResponseDataTypes = None,
        status_code: int = status.HTTP_200_OK,
        headers: dict | None = None,
        pagination: Pagination | None = None,
        set_cookies: list[Cookie] | None = None
    ):
        """
        :param data: should be an instance of ResponseDataTypes
        :param status_code: should be int
        :param headers: should be dict of headers
        :param pagination: an instance of Pagination or None
            The `pagination.template()` method will be used
        :param set_cookies: list of cookies you want to set on the client
            Set the `max_age` to `0` if you want to delete a cookie
        """
        headers = headers or {}
        self.pagination: Pagination | None = pagination
        if isinstance(data, Cursor):
            data = list(data)
        self.initial_data = data
        self.data = self.prepare_data(data=data)
        self.status_code = self.check_status_code(status_code=status_code)
        self.cookies = None
        if set_cookies:
            c = cookies.SimpleCookie()
            for cookie in set_cookies:
                c[cookie.key] = cookie.value
                c[cookie.key]['path'] = cookie.path
                c[cookie.key]['secure'] = cookie.secure
                c[cookie.key]['httponly'] = cookie.httponly
                c[cookie.key]['samesite'] = cookie.samesite
                if cookie.domain is not None:
                    c[cookie.key]['domain'] = cookie.domain
                if cookie.max_age is not None:
                    c[cookie.key]['max-age'] = cookie.max_age
            self.cookies = [(b'Set-Cookie', cookie.OutputString().encode()) for cookie in c.values()]
        self.headers = headers

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data

        if self.data is None:
            return b''
        return json.dumps(self.data)

    @property
    def headers(self) -> dict:
        return {
            'Content-Type': self.content_type,
            'Content-Length': len(self.body),
            'Access-Control-Allow-Origin': '*',
        } | self._headers

    @property
    def bytes_headers(self) -> list[tuple[bytes]]:
        result = [(k.encode(), str(v).encode()) for k, v in (self.headers or {}).items()]
        if self.cookies:
            result.extend(self.cookies)
        return result

    @headers.setter
    def headers(self, headers: dict):
        self._headers = headers

    @classmethod
    def prepare_data(cls, data: Any):
        """Make sure the response data is only ResponseDataTypes or Iterable of ResponseDataTypes"""
        if isinstance(data, (int | float | str | bool | bytes | NoneType)):
            return data

        elif isinstance(data, dict):
            return {key: cls.prepare_data(value) for key, value in data.items()}

        elif issubclass(type(data), BaseModel):
            return data.model_dump()

        elif isinstance(data, IterableDataTypes):
            return [cls.prepare_data(d) for d in data]

        else:
            msg = f'Invalid Response Type: {type(data)}'
            raise TypeError(msg)

    @classmethod
    def check_status_code(cls, status_code: Any):
        if not isinstance(status_code, int):
            error = f'Response `status_code` Should Be `int`. (`{status_code}` is {type(status_code)})'
            raise TypeError(error)
        return status_code

    async def send_headers(self, send, /):
        await send({'type': 'http.response.start', 'status': self.status_code, 'headers': self.bytes_headers})

    async def send_body(self, send, receive, /):
        await send({'type': 'http.response.body', 'body': self.body, 'more_body': False})

    async def send(self, send, receive, /):
        await self.send_headers(send)
        await self.send_body(send, receive)

    def __str__(self):
        if len(data := str(self.data)) > 30:
            data = f'{data:.27}...'
        return f'Response(status_code={self.status_code}, data={data})'

    __repr__ = __str__


class StreamingResponse(Response):
    content_type = 'application/octet-stream'

    def __init__(self, *args, **kwargs):
        self.connection_closed = False
        super().__init__(*args, **kwargs)

    async def listen_to_disconnection(self, receive):
        message = await receive()
        if message['type'] == 'http.disconnect':
            self.connection_closed = True

    def prepare_data(self, data: any) -> AsyncGenerator:
        if isinstance(data, AsyncGenerator):
            return data
        elif isinstance(data, Generator):
            return to_async_generator(data)
        msg = f'Invalid Response Type: {type(data)}'
        raise TypeError(msg)

    @property
    def headers(self) -> dict:
        return {
            'Content-Type': self.content_type,
            'Access-Control-Allow-Origin': '*',
        } | self._headers

    @headers.setter
    def headers(self, headers: dict):
        self._headers = headers

    @property
    async def body(self) -> AsyncGenerator:
        async for chunk in self.data:
            if isinstance(chunk, bytes):
                yield chunk
            elif chunk is None:
                yield b''
            else:
                yield json.dumps(chunk)

    async def send_body(self, send, receive, /):
        asyncio.create_task(self.listen_to_disconnection(receive))
        async for chunk in self.body:
            if self.connection_closed:
                break
            await send({'type': 'http.response.body', 'body': chunk, 'more_body': True})
        else:
            await send({'type': 'http.response.body', 'body': b'', 'more_body': False})


class HTMLResponse(Response):
    content_type = 'text/html; charset=utf-8'

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        return self.data.encode()


class PlainTextResponse(Response):
    content_type = 'text/plain; charset=utf-8'

    @property
    def body(self) -> bytes:
        if isinstance(self.data, bytes):
            return self.data
        return self.data.encode()


class TemplateResponse(HTMLResponse):
    """
    You may want to declare `TEMPLATES_DIR` in your configs

    Example:
        TEMPLATES_DIR = 'templates/'
            or
        TEMPLATES_DIR = '.'
    """
    def __init__(
        self,
        source: str | LiteralString | NoneType = None,
        name: str | NoneType = None,
        context: dict | NoneType = None,
        headers: dict | NoneType = None,
        status_code: int = status.HTTP_200_OK,
    ):
        """
        :param source: should be a string
        :param name: name of the template file (should be with its extension, e.g. index.html)
        :param context: should be dict of items
        :param headers: should be dict of headers
        :param status_code: should be int
        """
        if name:
            template = config.JINJA_ENVIRONMENT.get_template(name=name)
        else:
            template = config.JINJA_ENVIRONMENT.from_string(source=source)
        super().__init__(
            data=template.render(context or {}),
            headers=headers,
            status_code=status_code,
        )


class RedirectResponse(Response):
    def __init__(
        self,
        url: str,
        headers: dict | None = None,
        status_code: int = status.HTTP_307_TEMPORARY_REDIRECT,
        set_cookies: list[Cookie] | None = None
    ):
        headers = headers or {}
        headers['Location'] = url
        super().__init__(
            headers=headers,
            status_code=status_code,
            set_cookies=set_cookies,
        )



import logging
from time import perf_counter
from typing import Literal

from panther.base_request import BaseRequest
from panther.configs import config

logger = logging.getLogger('monitoring')


class Monitoring:
    """
    Create Log Message Like Below:
    date_time | method | path | ip:port | response_time(seconds) | status
    """
    def __init__(self, is_ws: bool = False):
        self.is_ws = is_ws

    async def before(self, request: BaseRequest):
        if config.MONITORING:
            ip, port = request.client

            if self.is_ws:
                method = 'WS'
            else:
                method = request.scope['method']

            self.log = f'{method} | {request.path} | {ip}:{port}'
            self.start_time = perf_counter()

    async def after(self, status: int | Literal['Accepted', 'Rejected', 'Closed'], /):
        if config.MONITORING:
            response_time = perf_counter() - self.start_time  # Seconds
            logger.info(f'{self.log} | {response_time} | {status}')



import logging

from panther import status
from panther.app import API, GenericAPI
from panther.authentications import JWTAuthentication, CookieJWTAuthentication
from panther.configs import config
from panther.db.models import BaseUser
from panther.exceptions import RedirectAPIError, AuthenticationAPIError
from panther.panel.middlewares import RedirectToSlashMiddleware
from panther.panel.utils import get_models, clean_model_schema
from panther.permissions import BasePermission
from panther.request import Request
from panther.response import TemplateResponse, Response, Cookie, RedirectResponse

logger = logging.getLogger('panther')


class AdminPanelPermission(BasePermission):
    @classmethod
    async def authorization(cls, request: Request) -> bool:
        try:  # We don't want to set AUTHENTICATION class, so we have to use permission classes
            await CookieJWTAuthentication.authentication(request=request)
            return True
        except AuthenticationAPIError:
            raise RedirectAPIError(url=f'login?redirect_to={request.path}')


class LoginView(GenericAPI):
    middlewares = [RedirectToSlashMiddleware]

    def get(self, request: Request):
        return TemplateResponse(name='login.html')

    async def post(self, request: Request):
        user: BaseUser = await config.USER_MODEL.find_one({config.USER_MODEL.USERNAME_FIELD: request.data['username']})
        if user is None:
            logger.debug('User not found.')
            return TemplateResponse(
                name='login.html',
                status_code=status.HTTP_400_BAD_REQUEST,
                context={'error': 'Authentication Error'},
            )
        if user.check_password(password=request.data['password']) is False:
            logger.debug('Password is incorrect.')
            return TemplateResponse(
                name='login.html',
                status_code=status.HTTP_400_BAD_REQUEST,
                context={'error': 'Authentication Error'},
            )
        tokens = JWTAuthentication.login(user.id)
        return RedirectResponse(
            url=request.query_params.get('redirect_to', '..'),
            status_code=status.HTTP_302_FOUND,
            set_cookies=[
                Cookie(
                    key='access_token',
                    value=tokens['access_token'],
                    max_age=config.JWT_CONFIG.life_time
                ),
                Cookie(
                    key='refresh_token',
                    value=tokens['refresh_token'],
                    max_age=config.JWT_CONFIG.refresh_life_time
                )
            ]
        )


class HomeView(GenericAPI):
    permissions = [AdminPanelPermission]

    def get(self):
        return TemplateResponse(name='home.html', context={'tables': get_models()})


class TableView(GenericAPI):
    permissions = [AdminPanelPermission]
    middlewares = [RedirectToSlashMiddleware]

    async def get(self, request: Request, index: int):
        model = config.MODELS[index]
        if data := await model.find():
            data = data
        else:
            data = []

        return TemplateResponse(
            name='table.html',
            context={
                'fields': clean_model_schema(model.schema()),
                'tables': get_models(),
                'records': Response.prepare_data(data),
            }
        )


class CreateView(GenericAPI):
    permissions = [AdminPanelPermission]
    middlewares = [RedirectToSlashMiddleware]

    async def get(self, request: Request, index: int):
        model = config.MODELS[index]
        return TemplateResponse(
            name='create.html',
            context={
                'fields': clean_model_schema(model.schema()),
                'tables': get_models(),
            }
        )

    async def post(self, request: Request, index: int):
        model = config.MODELS[index]
        validated_data = API.validate_input(model=model, request=request)
        instance = await model.insert_one(validated_data.model_dump())
        if issubclass(model, BaseUser):
            await instance.set_password(password=instance.password)
        return instance


class DetailView(GenericAPI):
    permissions = [AdminPanelPermission]
    middlewares = [RedirectToSlashMiddleware]

    async def get(self, index: int, document_id: str):
        model = config.MODELS[index]
        obj = await model.find_one_or_raise(id=document_id)
        return TemplateResponse(
            name='detail.html',
            context={
                'fields': clean_model_schema(model.schema()),
                'data': obj.model_dump()
            }
        )

    async def put(self, request: Request, index: int, document_id: str):
        model = config.MODELS[index]
        validated_data = API.validate_input(model=model, request=request)
        return await model.update_one({'id': document_id}, validated_data.model_dump())

    async def delete(self, index: int, document_id: str):
        model = config.MODELS[index]
        await model.delete_one(id=document_id)
        return Response(status_code=status.HTTP_204_NO_CONTENT)



import functools
import logging
import traceback
import typing
from datetime import timedelta
from typing import Literal, Callable

from orjson import JSONDecodeError
from pydantic import ValidationError, BaseModel

from panther._utils import is_function_async
from panther.caching import (
    get_response_from_cache,
    set_response_in_cache,
    get_throttling_from_cache,
    increment_throttling_in_cache
)
from panther.configs import config
from panther.exceptions import (
    APIError,
    AuthorizationAPIError,
    JSONDecodeAPIError,
    MethodNotAllowedAPIError,
    ThrottlingAPIError,
    BadRequestAPIError
)
from panther.exceptions import PantherError
from panther.middlewares import HTTPMiddleware
from panther.openapi import OutputSchema
from panther.permissions import BasePermission
from panther.request import Request
from panther.response import Response
from panther.serializer import ModelSerializer
from panther.throttling import Throttling

__all__ = ('API', 'GenericAPI')

logger = logging.getLogger('panther')


class API:
    """
    input_model: The `request.data` will be validated with this attribute, It will raise an
        `panther.exceptions.BadRequestAPIError` or put the validated data in the `request.validated_data`.
    output_schema: This attribute only used in creation of OpenAPI scheme which is available in `panther.openapi.urls`
        You may want to add its `url` to your urls.
    auth: It will authenticate the user with header of its request or raise an
        `panther.exceptions.AuthenticationAPIError`.
    permissions: List of permissions that will be called sequentially after authentication to authorize the user.
    throttling: It will limit the users' request on a specific (time-bucket, path)
    cache: Response of the request will be cached.
    cache_exp_time: Specify the expiry time of the cache. (default is `config.DEFAULT_CACHE_EXP`)
    methods: Specify the allowed methods.
    middlewares: These middlewares have inner priority than global middlewares.
    """
    func: Callable

    def __init__(
        self,
        *,
        methods: list[Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE']] | None = None,
        input_model: type[ModelSerializer] | type[BaseModel] | None = None,
        output_model: type[BaseModel] | None = None,
        output_schema: OutputSchema | None = None,
        auth: bool = False,
        permissions: list[BasePermission] | None = None,
        throttling: Throttling | None = None,
        cache: bool = False,
        cache_exp_time: timedelta | int | None = None,
        middlewares: list[HTTPMiddleware] | None = None,
    ):
        self.methods = {m.upper() for m in methods} if methods else None
        self.input_model = input_model
        self.output_schema = output_schema
        self.auth = auth
        self.permissions = permissions or []
        self.throttling = throttling
        self.cache = cache
        self.cache_exp_time = cache_exp_time
        self.middlewares: list[HTTPMiddleware] | None = middlewares
        self.request: Request | None = None
        if output_model:
            deprecation_message = (
                    traceback.format_stack(limit=2)[0] +
                    '\nThe `output_model` argument has been removed in Panther v5 and is no longer available.'
                    '\nPlease update your code to use the new approach. More info: '
                    'https://pantherpy.github.io/open_api/'
            )
            raise PantherError(deprecation_message)

    def __call__(self, func):
        self.func = func

        @functools.wraps(func)
        async def wrapper(request: Request) -> Response:
            chained_func = self.handle_endpoint
            if self.middlewares:
                for middleware in reversed(self.middlewares):
                    chained_func = middleware(chained_func)
            return await chained_func(request=request)

        # Store attributes on the function, so have the same behaviour as class-based (useful in `openapi.view.OpenAPI`)
        wrapper.auth = self.auth
        wrapper.methods = self.methods
        wrapper.permissions = self.permissions
        wrapper.input_model = self.input_model
        wrapper.output_schema = self.output_schema
        return wrapper

    async def handle_endpoint(self, request: Request) -> Response:
        self.request = request

        # 0. Preflight
        if self.request.method == 'OPTIONS':
            return self.options()

        # 1. Check Method
        if self.methods and self.request.method not in self.methods:
            raise MethodNotAllowedAPIError

        # 2. Authentication
        await self.handle_authentication()

        # 3. Permissions
        await self.handle_permission()

        # 4. Throttling
        await self.handle_throttling()

        # 5. Validate Input
        if self.request.method in {'POST', 'PUT', 'PATCH'}:
            self.handle_input_validation()

        # 6. Get Cached Response
        if self.cache and self.request.method == 'GET':
            if cached := await get_response_from_cache(request=self.request, cache_exp_time=self.cache_exp_time):
                return Response(data=cached.data, headers=cached.headers, status_code=cached.status_code)

        # 7. Put PathVariables and Request(If User Wants It) In kwargs
        kwargs = self.request.clean_parameters(self.func)

        # 8. Call Endpoint
        if is_function_async(self.func):
            response = await self.func(**kwargs)
        else:
            response = self.func(**kwargs)

        # 9. Clean Response
        if not isinstance(response, Response):
            response = Response(data=response)
        if response.pagination:
            response.data = await response.pagination.template(response.data)

        # 10. Set New Response To Cache
        if self.cache and self.request.method == 'GET':
            await set_response_in_cache(request=self.request, response=response, cache_exp_time=self.cache_exp_time)

        # 11. Warning CacheExpTime
        if self.cache_exp_time and self.cache is False:
            logger.warning('"cache_exp_time" won\'t work while "cache" is False')

        return response

    async def handle_authentication(self) -> None:
        if self.auth:
            if not config.AUTHENTICATION:
                logger.critical('"AUTHENTICATION" has not been set in configs')
                raise APIError
            self.request.user = await config.AUTHENTICATION.authentication(self.request)

    async def handle_throttling(self) -> None:
        if throttling := self.throttling or config.THROTTLING:
            if await get_throttling_from_cache(self.request, duration=throttling.duration) + 1 > throttling.rate:
                raise ThrottlingAPIError

            await increment_throttling_in_cache(self.request, duration=throttling.duration)

    async def handle_permission(self) -> None:
        for perm in self.permissions:
            if type(perm.authorization).__name__ != 'method':
                logger.error(f'{perm.__name__}.authorization should be "classmethod"')
                raise AuthorizationAPIError
            if await perm.authorization(self.request) is False:
                raise AuthorizationAPIError

    def handle_input_validation(self):
        if self.input_model:
            self.request.validated_data = self.validate_input(model=self.input_model, request=self.request)

    @classmethod
    def options(cls):
        headers = {
            'Access-Control-Allow-Methods': 'DELETE, GET, PATCH, POST, PUT, OPTIONS, HEAD',
            'Access-Control-Allow-Headers': 'Accept, Authorization, User-Agent, Content-Type',
        }
        return Response(headers=headers)

    @classmethod
    def validate_input(cls, model, request: Request):
        if isinstance(request.data, bytes):
            raise BadRequestAPIError(detail='Content-Type is not valid')
        if request.data is None:
            raise BadRequestAPIError(detail='Request body is required')
        try:
            # `request` will be ignored in regular `BaseModel`
            return model(**request.data, request=request)
        except ValidationError as validation_error:
            error = {'.'.join(str(loc) for loc in e['loc']): e['msg'] for e in validation_error.errors()}
            raise BadRequestAPIError(detail=error)
        except JSONDecodeError:
            raise JSONDecodeAPIError


class MetaGenericAPI(type):
    def __new__(
            cls,
            cls_name: str,
            bases: tuple[type[typing.Any], ...],
            namespace: dict[str, typing.Any],
            **kwargs
    ):
        if cls_name == 'GenericAPI':
            return super().__new__(cls, cls_name, bases, namespace)
        if 'output_model' in namespace:
            deprecation_message = (
                    traceback.format_stack(limit=2)[0] +
                    '\nThe `output_model` argument has been removed in Panther v5 and is no longer available.'
                    '\nPlease update your code to use the new approach. More info: '
                    'https://pantherpy.github.io/open_api/'
            )
            raise PantherError(deprecation_message)
        return super().__new__(cls, cls_name, bases, namespace)


class GenericAPI(metaclass=MetaGenericAPI):
    """
    Check out the documentation of `panther.app.API()`.
    """
    input_model: type[ModelSerializer] | type[BaseModel] | None = None
    output_schema: OutputSchema | None = None
    auth: bool = False
    permissions: list | None = None
    throttling: Throttling | None = None
    cache: bool = False
    cache_exp_time: timedelta | int | None = None
    middlewares: list[HTTPMiddleware] | None = None

    async def get(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def post(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def put(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def patch(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def delete(self, *args, **kwargs):
        raise MethodNotAllowedAPIError

    async def call_method(self, request: Request):
        match request.method:
            case 'GET':
                func = self.get
            case 'POST':
                func = self.post
            case 'PUT':
                func = self.put
            case 'PATCH':
                func = self.patch
            case 'DELETE':
                func = self.delete
            case 'OPTIONS':
                func = API.options
            case _:
                raise MethodNotAllowedAPIError

        return await API(
            input_model=self.input_model,
            output_schema=self.output_schema,
            auth=self.auth,
            permissions=self.permissions,
            throttling=self.throttling,
            cache=self.cache,
            cache_exp_time=self.cache_exp_time,
            middlewares=self.middlewares,
        )(func)(request=request)



import asyncio
import logging

from panther._utils import is_function_async
from panther.configs import config

logger = logging.getLogger('panther')


class Event:
    @staticmethod
    def startup(func):
        config.STARTUPS.append(func)

        def wrapper():
            return func()
        return wrapper

    @staticmethod
    def shutdown(func):
        config.SHUTDOWNS.append(func)

        def wrapper():
            return func()
        return wrapper

    @staticmethod
    async def run_startups():
        for func in config.STARTUPS:
            try:
                if is_function_async(func):
                    await func()
                else:
                    func()
            except Exception as e:
                logger.error(f'{func.__name__}() startup event got error: {e}')

    @staticmethod
    def run_shutdowns():
        for func in config.SHUTDOWNS:
            if is_function_async(func):
                try:
                    asyncio.run(func())
                except ModuleNotFoundError:
                    # Error: import of asyncio halted; None in sys.modules
                    #   And as I figured it out, it only happens when we are running with
                    #   gunicorn and Uvicorn workers (-k uvicorn.workers.UvicornWorker)
                    pass
            else:
                func()



from panther.panel.views import TableView, CreateView, LoginView, DetailView, HomeView

urls = {
    # '': models_api,
    # '<index>/': documents_api,
    # '<index>/<document_id>/': single_document_api,
    # 'health': healthcheck_api,
    '': HomeView,
    '<index>/': TableView,
    '<index>/create/': CreateView,
    'login/': LoginView,
    '<index>/<document_id>/': DetailView,
}
